import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.List;

import javax.swing.BoxLayout;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTabbedPane;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.JToggleButton;
import javax.swing.text.JTextComponent;


//TODO: For Version 2:
// Incorporate More control over totaling (taxes? subtotals? etc)
// Better Tax layout
// Preview Invoice?
// More Customization


public class InvoiceCreatorGUI extends JFrame implements ActionListener, FocusListener 
{
	// ------------------------------------------------------------------------
	// CLASS VARIABLES ******************************************************** 
	private String customPageCode;
	
	private JPanel invoicePanel;
	private JPanel listOfCalendars;
	
	private List<InvoiceEntry> invoices;
	
	// the rate to apply to all entries
	private JTextField hourlyRateField;
	
	// this stores a link to the Google Calendar URL
	private JTextField calendarLink;
	
	// this stores a list of CalendarReaderButtons, each reference to the CalendarReader queries the Google Server
	private ArrayList<CalendarReader> calendars;
	
	// this button, when clicked, causes the CalendarReader to activate
	private JButton calendarQueryButton;
	
	// this button, when clicked, dumps the invoice to a file
	private JButton saveInvoiceButton;
	
	private JButton selectVisibleButton;
	private JButton deSelectVisibleButton;
	
	private JToggleButton hideEntriesBeforeButton;
	private JToggleButton hideEntriesAfterButton;
	
	private JTextField startDateField;
	private JTextField endDateField;
	
	// these checkboxes are used by advancedSettingsPanel
	private JCheckBox sortByCalendarCheckBox;
	private JCheckBox calendarTitleCheckBox;
	private JCheckBox calendarCaptionCheckBox;
	private JCheckBox calendarTotalTimeCheckBox;
	private JCheckBox entryTitleCheckBox;
	private JCheckBox entryDescriptionCheckBox;
	private JCheckBox entryTimeStampCheckBox;
	private JCheckBox entryLocationCheckBox;
	private JCheckBox entryTotalTimeCheckBox;
	private JButton applyAdvancedChangesButton;
	
	// these fields are used by the Header/Footer panel
	private JTextField myLogo;
	private JTextField myName;
	private JTextField myEmail;
	private JTextField myPhone;
	private JTextField myOther;
	private JTextField invoiceDate;
	private JTextField clientName;
	private JTextField clientAddress1;
	private JTextField clientAddress2;
	private JTextField clientAddress3;
	private JTextArea headerComment;
	private JTextArea footerComment;
	private JTextField myAddress1;
	private JTextField myAddress2;
	private JTextField myAddress3;
	private JButton advancedHeaderFooterButton;
	private JCheckBox useAdvancedHeaderFooterCheckBox;
	
	private JLabel invoiceInterfaceHours;
	
	// the Advanced Header Footer object is used to more finely customize your
	// header and footer - it is a seperate popup that communicates with the main GUI
	AdvancedHeaderFooter advHeaderFooter;
	
	
	// ------------------------------------------------------------------------
	// SINGLETON PATTERN ******************************************************
	private static InvoiceCreatorGUI instance;
	
	private InvoiceCreatorGUI()
	{		
		// intitiate the private variables
		calendarLink = new JTextField();
		hourlyRateField = new JTextField();
		
		// start off with 0 invoice entries loaded
		invoices = new ArrayList<InvoiceEntry>();
		calendars = new ArrayList<CalendarReader>();
		
		// the basic Header Footer modifier fields are declared here
		myLogo = new JTextField("link to custom logo");
		myLogo.addFocusListener(this);
		myName = new JTextField("my full name");
		myName.addFocusListener(this);
		myEmail = new JTextField("my email address");
		myEmail.addFocusListener(this);
		myPhone = new JTextField("my phone number");
		myPhone.addFocusListener(this);
		myOther = new JTextField("other field - optional");
		myOther.addFocusListener(this);
		invoiceDate = new JTextField("date of issue");
		invoiceDate.addFocusListener(this);
		clientName = new JTextField("client name");
		clientName.addFocusListener(this);
		clientAddress1 = new JTextField("client address line 1");
		clientAddress1.addFocusListener(this);
		clientAddress2 = new JTextField("client address line 2");
		clientAddress2.addFocusListener(this);
		clientAddress3 = new JTextField("client address line 3");
		clientAddress3.addFocusListener(this);
		headerComment = new JTextArea("final header comment - optional");
		headerComment.addFocusListener(this);
		footerComment = new JTextArea("footer comment - optional");
		footerComment.addFocusListener(this);
		myAddress1 = new JTextField("my address line 1");
		myAddress1.addFocusListener(this);
		myAddress2 = new JTextField("my address line 2");
		myAddress2.addFocusListener(this);
		myAddress3 = new JTextField("my address line 3");
		myAddress3.addFocusListener(this);
		advancedHeaderFooterButton = new JButton("Advanced Header/Footer");
		advancedHeaderFooterButton.addActionListener(this);
		advancedHeaderFooterButton.setToolTipText("Opens a window to customize Header/Footer. Recommended for advanced users only");
		useAdvancedHeaderFooterCheckBox = new JCheckBox("Use Advanced Header/Footer", false);
		useAdvancedHeaderFooterCheckBox.setToolTipText("If checked, the information in the advanced Header/Footer popup will be used.");
		
		// set the text in the saveInvoiceButton
		saveInvoiceButton = new JButton("Save to File");
		saveInvoiceButton.addActionListener(this);
		
		// set the text in the calendarQueryButton
		calendarQueryButton = new JButton("Add Google Calendar");
		calendarQueryButton.addActionListener(this);
		
		// the invoice panel displays the complete list of all loaded invoice entries
		// it is oriented on the right side of the window
		invoicePanel = new JPanel();
		invoicePanel.setLayout(new BoxLayout(invoicePanel,BoxLayout.Y_AXIS));
		
		// default explain that not invoices have been loaded
		invoicePanel.add(new JLabel("No Invoices Loaded..."));
		
		JScrollPane invoiceScroller = new JScrollPane(invoicePanel, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		invoiceScroller.getVerticalScrollBar().setUnitIncrement(60);
		invoiceScroller.getVerticalScrollBar().setBlockIncrement(180);
		
		invoiceInterfaceHours = new JLabel("0");
		
		JPanel invoiceStats = new JPanel(new GridLayout(1,2));
		invoiceStats.add(new JLabel("Total Hours: "));
		invoiceStats.add(invoiceInterfaceHours);
		
		JPanel invoiceInterfacePanel = new JPanel(new BorderLayout());
		invoiceInterfacePanel.add(invoiceScroller, BorderLayout.CENTER);
		invoiceInterfacePanel.add(invoiceStats, BorderLayout.SOUTH);

		// the links and labels panel stores the Calendar Link, labels, and the
		// calendar query button
		JPanel linksAndLabels = new JPanel(new GridLayout(3,1));
		((GridLayout)linksAndLabels.getLayout()).setVgap(5);
		linksAndLabels.add(new JLabel("Link to a Google Calendar Feed:"));
		linksAndLabels.add(calendarLink);
		linksAndLabels.add(calendarQueryButton);
		
		// the contentPanel stores all the customizing input fields for the header and footer
		// and also a button to open a more advanced method of customization
		JPanel contentPanel = new JPanel();
		contentPanel.setLayout(new BoxLayout(contentPanel, BoxLayout.PAGE_AXIS));
		contentPanel.add(new JLabel("*Unused fields should have their content deleted before creating your Invoice"));
		contentPanel.add(new JLabel("---------- HEADER ----------"));
		contentPanel.add(myLogo);
		contentPanel.add(myName);
		contentPanel.add(myEmail);
		contentPanel.add(myPhone);
		contentPanel.add(myOther);
		contentPanel.add(invoiceDate);
		contentPanel.add(clientName);
		contentPanel.add(clientAddress1);
		contentPanel.add(clientAddress2);
		contentPanel.add(clientAddress3);
		contentPanel.add(headerComment);
		contentPanel.add(new JLabel(" "));
		contentPanel.add(new JLabel("---------- FOOTER ----------"));
		contentPanel.add(footerComment);
		contentPanel.add(myAddress1);
		contentPanel.add(myAddress2);
		contentPanel.add(myAddress3);
		contentPanel.add(new JLabel(" "));
		contentPanel.add(advancedHeaderFooterButton);
		contentPanel.add(useAdvancedHeaderFooterCheckBox);

		JScrollPane contentPanelScroller = new JScrollPane(contentPanel, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_AS_NEEDED);
		contentPanelScroller.getVerticalScrollBar().setUnitIncrement(60);
		contentPanelScroller.getVerticalScrollBar().setBlockIncrement(180);
		
		// the final panel has the hourly rate (label and field) and the save invoice button
		JPanel finalPanel = new JPanel();
		finalPanel.setLayout(new BoxLayout(finalPanel, BoxLayout.X_AXIS));
		finalPanel.add(new JLabel("Hourly Rate:  "));
		finalPanel.add(hourlyRateField);
		finalPanel.add(new JLabel(""));
		finalPanel.add(saveInvoiceButton);
		finalPanel.setMaximumSize(new Dimension(100, 100));
		
		// This block is comprised of all Date Range Panel objects
		// this includes:
		//   -- Hide Entries BEFORE Button
		//   -- Hide Entries AFTER Button
		//   -- The BEFORE Date Field
		//   -- The AFTER Date Field
		//   -- the SELECT Visible Button 
		//   -- the DESELECT Visible Button
		//-----------------------------------------------
		
		hideEntriesBeforeButton = new JToggleButton("Hide Entries Before: ");
		hideEntriesBeforeButton.addActionListener(this);
		
		hideEntriesAfterButton = new JToggleButton("Hide Entries After: ");
		hideEntriesAfterButton.addActionListener(this);
		
		JPanel dateRangePanel = new JPanel(new GridLayout(4,2));
		dateRangePanel.add(hideEntriesBeforeButton);
		
		startDateField = new JTextField("");
		startDateField.setToolTipText("Format of 'dd-MM-yy'");
		dateRangePanel.add(startDateField);
		
		dateRangePanel.add(hideEntriesAfterButton);
		
		endDateField = new JTextField("");
		endDateField.setToolTipText("Format of 'dd-MM-yy'");
		dateRangePanel.add(endDateField);
		
		selectVisibleButton = new JButton("Select Visible");
		selectVisibleButton.addActionListener(this);
		
		deSelectVisibleButton = new JButton("DeSelect Visible");
		deSelectVisibleButton.addActionListener(this);
		
		dateRangePanel.add(selectVisibleButton);
		dateRangePanel.add(deSelectVisibleButton);
		dateRangePanel.add(new JLabel(" "));
		dateRangePanel.add(new JLabel(" "));
		
		//-----------------------------------------------
		
		listOfCalendars = new JPanel();
		listOfCalendars.setLayout(new BoxLayout(listOfCalendars,BoxLayout.Y_AXIS));
		listOfCalendars.add(new JLabel("-- NONE --"));
		
		JPanel calendarListHolder = new JPanel();
		calendarListHolder.setLayout(new BorderLayout());
		calendarListHolder.add(new JLabel("List of Calendars:"), BorderLayout.NORTH);
		calendarListHolder.add(listOfCalendars, BorderLayout.CENTER);
		
		JScrollPane calendarScroller = new JScrollPane(calendarListHolder, JScrollPane.VERTICAL_SCROLLBAR_AS_NEEDED, JScrollPane.HORIZONTAL_SCROLLBAR_NEVER);
		calendarScroller.getVerticalScrollBar().setUnitIncrement(60);
		calendarScroller.getVerticalScrollBar().setBlockIncrement(180);
		
		JPanel settingsPanelFotter = new JPanel();
		settingsPanelFotter.setLayout(new BorderLayout());
		settingsPanelFotter.add(dateRangePanel, BorderLayout.CENTER);
		settingsPanelFotter.add(finalPanel, BorderLayout.SOUTH);
		
		// the settings panel stores basic the user functionality, it is oriented on the
		// left side of the window. Inside this is the content panel
		JPanel settingsPanel = new JPanel();
		settingsPanel.setLayout(new BorderLayout());
		settingsPanel.add(linksAndLabels, BorderLayout.NORTH);
		settingsPanel.add(settingsPanelFotter, BorderLayout.SOUTH);
		settingsPanel.add(calendarScroller, BorderLayout.CENTER);
		
		// the advanced settings checkboxes are declared here
		sortByCalendarCheckBox = new JCheckBox("Sort Invoice by Calendar", false);
		sortByCalendarCheckBox.setToolTipText("If checked, invoices will be grouped by the Calendars they came from");
		calendarTitleCheckBox = new JCheckBox("Calendar Title", true);
		calendarCaptionCheckBox = new JCheckBox("Calendar Description", false);
		calendarTotalTimeCheckBox = new JCheckBox("Total Time", true);
		entryTitleCheckBox = new JCheckBox("Entry Title", true);
		entryDescriptionCheckBox = new JCheckBox("Entry Description", false);
		entryTimeStampCheckBox = new JCheckBox("Entry Timestamp", true);
		entryLocationCheckBox = new JCheckBox("Entry Location", false);
		entryTotalTimeCheckBox = new JCheckBox("Entry Total Time", true);
		
		// the button to apply these changes is declared here
		applyAdvancedChangesButton = new JButton("Apply Changes");
		applyAdvancedChangesButton.setToolTipText("Will update the invoice to reflect the new formatting changes");
		applyAdvancedChangesButton.addActionListener(this);
		
		// the advancedSettings panel stores more complicated features like determining what information
		// is to be included from each Entry, date range inclusions, etc
		JPanel advancedSettingsPanel = new JPanel();
		advancedSettingsPanel.setLayout(new BoxLayout(advancedSettingsPanel,BoxLayout.Y_AXIS));
		advancedSettingsPanel.add(sortByCalendarCheckBox);
		advancedSettingsPanel.add(new JLabel(" "));
		advancedSettingsPanel.add(new JLabel("Once Per Calendar:"));
		advancedSettingsPanel.add(calendarTitleCheckBox);
		advancedSettingsPanel.add(calendarCaptionCheckBox);
		advancedSettingsPanel.add(calendarTotalTimeCheckBox);
		advancedSettingsPanel.add(new JLabel(" "));
		advancedSettingsPanel.add(new JLabel("For Each Entry:"));
		advancedSettingsPanel.add(entryTitleCheckBox);
		advancedSettingsPanel.add(entryLocationCheckBox);
		advancedSettingsPanel.add(entryTimeStampCheckBox);
		advancedSettingsPanel.add(entryDescriptionCheckBox);
		advancedSettingsPanel.add(entryTotalTimeCheckBox);
		advancedSettingsPanel.add(new JLabel(" "));
		advancedSettingsPanel.add(new JLabel(" "));
		advancedSettingsPanel.add(applyAdvancedChangesButton);
		
		// this tabbedpane separates the file formatting fields from the background
		// calendar loader and calculating fields
		JTabbedPane tabbedPane = new JTabbedPane();
		tabbedPane.addTab("Calendar Info", settingsPanel);
		tabbedPane.addTab("Header/Footer", contentPanelScroller);
		tabbedPane.addTab("Entry Formatting", advancedSettingsPanel);
		
		// maingrid splits the main window into left hand / right hand (right hand being the invoices)
		GridLayout mainGrid = new GridLayout(1,2);
		mainGrid.setHgap(10);
		
		// construct the frame window
		this.setLayout(mainGrid);
		this.add(tabbedPane);
		this.add(invoiceInterfacePanel);
		this.setSize(800, 600);
		this.setVisible(true);
	    this.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
	}
	
	// Simple Singleton pattern
	public static InvoiceCreatorGUI getInstance()
	{
		if (instance == null)
		{
			instance = new InvoiceCreatorGUI();
		}
		
		return instance;
	}
	
	// ------------------------------------------------------------------------
	// BACKGROUND FUNCTIONS ***************************************************
	// this function is called to handled action events created by buttons when they're
	// pressed by the user
	@Override
	public void actionPerformed(ActionEvent event) 
	{
		// get the button responsible for triggering the pressed event
		try
		{
			JButton button = (JButton) event.getSource();
			
			// if the button pressed was the saveInvoiceButton
			if(button.equals(saveInvoiceButton))
			{
				// create a File Chooser for the user to save the file
				JFileChooser fileChooser = new JFileChooser((new File(".")).getAbsolutePath());
				
				// if they approved the transaction
				if(fileChooser.showSaveDialog(null) == JFileChooser.APPROVE_OPTION)
				{
					// extract the filepath
					String filePath = fileChooser.getSelectedFile().getAbsolutePath();
					
					// if the path does not end with .html
					if(!filePath.endsWith(".html"))
					{
						// we will append it
						filePath=filePath + ".html";
					}
					
					// create the file
					File invoiceFile = new File(filePath);
					
					// this flag will be used to determine whether or not the file has been cleared for creation
					boolean createFileFlag = false;
					
					// if that file exists
					if(invoiceFile.exists())
					{
						// ask the user if they are willing to overwrite that file, if they say yes
						if(JOptionPane.showConfirmDialog(null, "That file already exists, would you like to overwrite it?") == JOptionPane.YES_OPTION)
						{
							// clear the file for creation
							createFileFlag = true;
						}
					}
					// if no file exists with that name and path
					else
					{
						// clear the file for creation
						createFileFlag = true;
					}
					
					// if the file is prepared to be created
					if(createFileFlag)
					{
						// get the completed invoice text
						String invoiceContent = getInvoiceContent();
						
						// write the invoice
						InvoiceWriter invoiceWriter = new InvoiceWriter(invoiceFile, invoiceContent);
						try 
						{
							invoiceWriter.write();
						} 
						catch (IOException e) 
						{
							JOptionPane.showMessageDialog(null, "Invoice encountered an error while saving, please try again.");
						}
					}
					// if the file was not cleared to be created
					else
					{
						// inform them the file could not be saved
						JOptionPane.showMessageDialog(null, "Invoice not saved! Please try again.");
					}
				}
				
			}
			// if the button pressed was the calendarQueryButton
			else if (button.equals(calendarQueryButton))
			{
				try 
				{
					// attempt to add a new Calendar to the list
					addCalendar();
					
					// notify the user that the Calendar was loaded successfully
					JOptionPane.showMessageDialog(null, "Calendar loaded successfully!");
					
					// clear the calendar entry field
					calendarLink.setText("");
				} 
				catch (MalformedURLException e) 
				{
					// this exception is caught when the user enters an invalid URL
					JOptionPane.showMessageDialog(null, "The URL you entered is invalid, please verfiy it's accuracy and try again.");
				} 
				catch (CalendarAlreadyLoadedError e)
				{
					// this exception is caught when the user enters a link to a Calendar already in use
					JOptionPane.showMessageDialog(null, e.getMessage());
				}
				catch (Exception e) 
				{
					// this exception is caught when there is an XML encoding error
					// will happen if they enter a valid URL that does not return an XML object
					JOptionPane.showMessageDialog(null, "There was an error reading the Calendar data. Please verfiy the link's accuracy and try again.");
					e.printStackTrace();
				}
			}
			// if the applyAdvancedChangesButton button pressed, apply the formatting changes
			else if(button.equals(applyAdvancedChangesButton))
			{
				// first determine the User's new format settings
				InvoiceEntryFormat ie = getUserInvoiceEntryFormat();
				
				for(int i=0; i<invoices.size(); i++)
				{
					// update the text in every entry using the user's preferred settings
					((InvoiceEntryButton)invoicePanel.getComponent(i)).refreshText(ie);
				}
			}
			// if the advancedHeaderFooterButton button was pressed, popup a header/footer object for advanced editing
			else if(button.equals(advancedHeaderFooterButton))
			{
				// if we have not yet initialized the GUI, create it
				// *Singleton Implementation not an option for the Object itself
				if(advHeaderFooter == null)
				{
					advHeaderFooter = new AdvancedHeaderFooter();
				}
				
				advHeaderFooter.popUp();
			}
			else if(button.equals(selectVisibleButton))
			{
				setSelectionForVisible(true);
			}
			else if(button.equals(deSelectVisibleButton)){
				setSelectionForVisible(false);
			}
		}
		catch(ClassCastException e)
		{
			JToggleButton button = (JToggleButton) event.getSource();
			
			if(button.equals(hideEntriesBeforeButton))
			{
				boolean flag = hideEntriesBeforeDate(!hideEntriesBeforeButton.isSelected());
				if(!flag)
				{
					JOptionPane.showMessageDialog(null, "You must enter a date in the format 'dd-MM-yy'");
					hideEntriesBeforeButton.setSelected(false);
				}
			}
			else if(button.equals(hideEntriesAfterButton))
			{
				boolean flag = hideEntriesAfterDate(!hideEntriesAfterButton.isSelected());
				if(!flag)
				{
					JOptionPane.showMessageDialog(null, "You must enter a date in the format 'dd-MM-yy'");
					hideEntriesAfterButton.setSelected(false);
				}
			}
		}
	}

	// this takes a visibility flag (boolean) and applies it as a visibility
	// setting for all invoices BEFORE the value stored in "startDateField"
	// it returns TRUE if the operation was successful and FALSE if it failed
	// to parse the date value
	private boolean hideEntriesBeforeDate(boolean flag) 
	{
		boolean retVal = true;
		
		String dateText = startDateField.getText();
		
		SimpleDateFormat formatter = new SimpleDateFormat("dd-MM-yy");
		try 
		{
			Date date = (Date)formatter.parse(dateText);
			for(int i=0; i<invoices.size(); i++)
			{
				InvoiceEntryButton thisEntry = (InvoiceEntryButton)invoicePanel.getComponent(i);
				
				if(date.compareTo(thisEntry.getEntry().getEndDate())>0)
				{
					thisEntry.makeVisible(flag);
				}
				else
				{
					break;
				}
			}
		} 
		catch (ParseException e) 
		{
			retVal = false;
		}
		
		return retVal;
	}

	// this takes a visibility flag (boolean) and applies it as a visibility
	// setting for all invoices AFTER the value stored in "endDateField"
	// it returns TRUE if the operation was successful and FALSE if it failed
	// to parse the date value
	private boolean hideEntriesAfterDate(boolean flag) 
	{
		boolean retVal = true;
		
		String dateText = endDateField.getText();
		
		SimpleDateFormat formatter = new SimpleDateFormat("dd-MM-yy");
		try 
		{
			Date date = (Date)formatter.parse(dateText);
			
			for(int i=invoices.size()-1; i>=0; i--)
			{
				InvoiceEntryButton thisEntry = (InvoiceEntryButton)invoicePanel.getComponent(i);
				
				if(date.compareTo(thisEntry.getEntry().getStartDate())<0)
				{
					thisEntry.makeVisible(flag);
				}
				else
				{
					break;
				}
			}
		} 
		catch (ParseException e) 
		{
			retVal = false;
		}
		
		return retVal;
	}

	// this method compiles the invoice content based on the demands of the user's preferences
	private String getInvoiceContent() 
	{
		String retVal = null;
		
		// if they wish to create the invoice using advanced settings...
		if(useAdvancedHeaderFooterCheckBox.isSelected())
		{
			// return the invoice constructed with advanced settings
			retVal = getInvoiceContentWithAdvanced();
			// JOptionPane.showMessageDialog(null, "CANNOT MAKE INVOICE WITH ADVANCED SETTINGS YET!");
		}
		// if they wish to create the invoice using the template and forms
		else
		{
			// return the invoice constructed with helpful forms
			retVal = getInvoiceContentWithForms();
		}
		
		// return the invoice
		return retVal;
	}

	// this function creates
	private String getInvoiceContentWithAdvanced() 
	{
		String retVal = new String();
		
		String text = advHeaderFooter.getText();
		
		String pivotString = text.substring(text.indexOf("id=\"invoices\""));
		int extraLength = pivotString.substring(0, pivotString.indexOf(">")).length()+1;
		
		// the insertIndex holds the location in the string where the invoiceFormat() should be placed
		int insertIndex = text.indexOf("id=\"invoices\"") + extraLength;
		
		// create the final value
		retVal = text.substring(0, insertIndex) + formatInvoices() + text.substring(insertIndex);
		
		return retVal;
	}

	// this function creates a templated invoice with their form field replacement content
	private String getInvoiceContentWithForms() 
	{
		String retVal = new String();
		
		String logoAddress = myLogo.getText();
		String servicerName = myName.getText();
		String servicerEmail = myEmail.getText();
		String servicerPhone = myPhone.getText();
		String servicerOther = myOther.getText();
		String servicerAddress1 = myAddress1.getText();
		String servicerAddress2 = myAddress2.getText();
		String servicerAddress3 = myAddress3.getText();
		String invoiceTimeStamp = invoiceDate.getText();
		String issuedName = clientName.getText();
		String issuedAddress1 = clientAddress1.getText();
		String issuedAddress2 = clientAddress2.getText();
		String issuedAddress3 = clientAddress3.getText();
		String headerCommentString = headerComment.getText();
		String footerCommentString = footerComment.getText();
		
		
		// headText is essentially the information inside the <head> tags
		String headText = "<html><head><style>html, body{text-align:left;font-family:\"Courier New\", Courier, monospace;background-color:#FFFFFF;color:#000000;margin:0;padding:0;line-height:20px;}.bold{font-weight:bold;}.italic{font-style:italic;}.large{font-size:16px;}.normal{font-size:12px;}.small{font-size:10px;}p{margin:0;padding:0;}div.header{width:576;}div.logo{background-image:url(" + logoAddress + ");background-position:top 0;background-size:contain;background-repeat:no-repeat;width:200px;height:100px;position:absolute;top:0px;}div.header_info{width:100%;height:auto;text-align:right;}div.client_info{text-align:center;width:100%;height:auto;}div.footer{width:576px;height:auto;text-align:center;}div.invoices{width:576px;height:auto;text-align:left;}.summation{margin-left:50%;}.entry{margin:0px 50px 50px 50px}.calendar_group{width:100%;height:auto;}.left{float:left;}.right{width:100%;text-align:right;}.calendar{background-image:url(images/bg_color.gif);background-repeat:repeat-y;}.calendar_divider{background-image:url(images/bg_color.gif);margin-top:20px;margin-bottom:20px;height:3px;}</style></head>";
		
		// preContent is the header information
		String preContent = "<body><div class=\"header\"><div class=\"logo\"></div><div class=\"header_info\"> <p id=\"username\" class=\"bold\">"+ servicerName 
							+"</p><p id=\"useremail\">" + servicerEmail
							+"</p><p id=\"userphone\">" + servicerPhone
							+"</p><p id=\"userother\">" + servicerOther
							+"</p></div>" 
							+"<div style=\"height:50px;\"></div>" 
							+"<div class=\"client_info\"><p id=\"issuedate\">" + invoiceTimeStamp
							+"<div style=\"height:50px;\"></div>" 
							+"</p><p id=\"clientname\" class=\"bold\">" + issuedName 
							+"</p><p id=\"clientad1\">" + issuedAddress1 
							+"</p><p id=\"clientad2\">" + issuedAddress2
							+"</p><p id=\"clientad3\">" + issuedAddress3
							+"<div style=\"height:50px;\"></div>" 
							+"</p><p id=\"headercomment\">" + headerCommentString + "</p></div></div>"
							+"<div style=\"height:50px;\"></div>";
		
		// invoiceContent is the actual invoice information
		String invoiceContent = "<div class=\"invoices\">"
							  +formatInvoices()
							  +"</div>";
		
		// postContent is the footer information
		String postContent = "<div style=\"height:50px;\"></div><div class=\"footer\"><p id=\"footercomment\">" + footerCommentString
						   +"</p><br /><p id=\"userad1\">" + servicerAddress1
						   +"</p><p id=\"userad2\">" + servicerAddress2
						   +"</p><p id=\"userad3\">"+ servicerAddress3
						   +"</p></div></body></html>";
		
		retVal = headText + preContent + invoiceContent + postContent;
		
		return retVal;
	}

	// this function returns an HTML string that has all the invoice information
	private String formatInvoices() 
	{
		String retVal = "";
		
		// get all the entries designated for inclusion in the final Invoice
		ArrayList<InvoiceEntry> selectedEntries = getSelectedEntries();
		try
		{
			// if the user wishes to split the invoice up by Calendar
			if(sortByCalendarCheckBox.isSelected())
			{
				// create an array that is split by Calendar types
				ArrayList<ArrayList<InvoiceEntry>> sortedSelectedEntries = splitByCalendar(selectedEntries);
				
				for(int i=0; i<sortedSelectedEntries.size(); i++)
				{
					retVal += "<div class=\"large bold\">" + calendars.get(i).getTitle() + "</div><div class=\"calendar\">";
					retVal += lumpEntries(sortedSelectedEntries.get(i));
					retVal += "</div>";
					
					// if there is another calendar to follow
					if(i<sortedSelectedEntries.size()-1)
					{
						// we add a space
						retVal += "<div class=\"calendar_divider\"></div>";
					}
				}
			}
			// if the user would rather lump all the entries together
			else
			{
				retVal = lumpEntries(selectedEntries);
			}
		}
		catch(NumberFormatException e)
		{
			JOptionPane.showMessageDialog(null, "You must enter a valid hourly rate!");
		}
		return retVal;
	}

	// this function takes a list of InvoiceEntries and formats them into HTML
	// that can be included on the final Invoice
	@SuppressWarnings("unchecked")
	private String lumpEntries(ArrayList<InvoiceEntry> selectedEntries) throws NumberFormatException
	{
		String retVal = "";
		double hours = 0;
		double rate = Double.parseDouble(hourlyRateField.getText());
		
		// first sort the list
		Collections.sort(selectedEntries);
		
		// for every entry in the list
		for(int i=0; i<selectedEntries.size(); i++)
		{
			retVal += selectedEntries.get(i).printHTML(getUserInvoiceEntryFormat());
			hours += selectedEntries.get(i).getHours();
		}
		
		retVal += "<div class=\"summation\"><div class=\"left\">TOTAL HOURS:</div><div class=\"right\">" + hours + "</div>";
		retVal += "<div class=\"left\">RATE:</div><div class=\"right\">$" + rate + "</div>";
		retVal += "<div class=\"left bold large\">TOTAL:</div><div class=\"right bold large\">$" + hours*rate + "</div></div>";
		
		return retVal;
	}

	// this function takes a list of entries and returns a list of a list of entries
	// each element in the greater list represents a calendar and the elements within
	// each element in the greater list represents an invoice entry belonging to that
	// calendar
	private ArrayList<ArrayList<InvoiceEntry>> splitByCalendar(List<InvoiceEntry> selectedEntries) 
	{
		// create an empty list
		ArrayList<ArrayList<InvoiceEntry>> retVal = new ArrayList<ArrayList<InvoiceEntry>>();
		
		// for every Calendar we have loaded
		for(int i=0; i<calendars.size(); i++)
		{
			// we will create a new (empty) list of invoice entries
			retVal.add(new ArrayList<InvoiceEntry>());
		}
		
		// for every entry that needs to be sorted into calendars
		for(InvoiceEntry entry : selectedEntries)
		{
			// first we get the location in the array of the calendar to whom
			// this entry belongs
			int calendarLocation = calendars.indexOf(entry.getParent());
			
			// we then add (push / append) this entry to the end of the list
			// matching the location of that calendar in the list of list<InvoiceEntry>
			retVal.get(calendarLocation).add(entry);
		}
		
		// return the list of lists
		return retVal;
	}

	// this function returns a list of InvoiceEntries corresponding to those InvoiceEntryButtons that 
	// the user selected to be included in the final invoice
	private ArrayList<InvoiceEntry> getSelectedEntries() 
	{
		ArrayList<InvoiceEntry> retVal = new ArrayList<InvoiceEntry>();
		
		// for every invoice,
		for(int i=0; i<invoices.size(); i++)
		{
			// get this invoice entry button
			InvoiceEntryButton thisEntry = (InvoiceEntryButton)invoicePanel.getComponent(i);
			
			// thisEntry is selected
			if(thisEntry.isSelected())
			{
				// add this entry to the list of entries
				retVal.add(thisEntry.getEntry());
			}
		}
		
		return retVal;
	}

	// this function adds another Calendar to the list of Calendars
	private void addCalendar() throws MalformedURLException, CalendarAlreadyLoadedError, DataNotLoaded, Exception
	{
		if(CalendarLoaded(calendarLink.getText()))
		{
			throw new CalendarAlreadyLoadedError();
		}
		// create the new calendar reader
		CalendarReader calendar = new CalendarReader(calendarLink.getText());
		
		// load the data at the address
		calendar.loadXML();
		
		// create the CalendarReaderButton to match
		calendars.add(calendar);
		
		// load the invoices
		System.out.println("LOADING INVOICES");
		loadInvoices();
		
		// update the complete list of CalendarReaders[Buttons]
		updateCalendarList();
	}
	
	// this function takes in a string (a URL) and checks to see if a Calendar has already
	// been loaded with that URL
	private boolean CalendarLoaded(String text) 
	{
		// retVal will store the result of the search through the calendar list
		boolean retVal = false;
	
		// for each loaded calendar
		for(CalendarReader calendar : calendars)
		{
			// check if it's URL matches the designated URL
			if(calendar.getURL().equals(text))
			{
				// if the two match, set retVal to true and exit
				retVal = true;
				break;
			}
		}
		
		// return the result of the search
		return retVal;
	}

	// this function reloads all the invoices using the CalendarReader's built in functionality
	// it also creates button representations of every InvoiceEntry object
	// TODO: A current bug with the way invoices are loaded is that any invoice selected before
	// a new calendar was loaded becomes reset back to unselected after the list is refreshed with new
	// invoices. A better functionality must be discovered that simply APPENDS new invoices without
	// disturbing existing ones
	@SuppressWarnings("unchecked")
	private void loadInvoices() throws DataNotLoaded
	{
		// remove all the current invoices
		invoices.clear();
		
		// for each CalendarReaderButton the GUI maintains
		for(int i=0; i<calendars.size(); i++)
		{
			// grab a reference to the actual CalendarReader
			CalendarReader calendar = calendars.get(i);
			
			// add the list of invoices loaded from the Calendar Reader to the growing total
			invoices.addAll(calendar.getInvoicesFromXML());
		}
		
		// sort the complete list of invoices
		Collections.sort(invoices);
		
		// create a new InvoicePanel to hold the results
		invoicePanel.removeAll();
		invoicePanel.setLayout(new GridLayout(invoices.size(),1));
		((GridLayout)invoicePanel.getLayout()).setVgap(4);
		
		InvoiceEntryFormat ie = getUserInvoiceEntryFormat();
		
		// for every invoice in the list
		for(InvoiceEntry entry : invoices)
		{
			// we create a new InvoiceEntryButton linked to this current entry
			InvoiceEntryButton thisButton = new InvoiceEntryButton(entry);
			
			// we set it's text based on the selected preferences of the User
			thisButton.refreshText(ie);
			
			// and then add that button to the panel of invoice entry GUI elements
			invoicePanel.add(thisButton);
		}
		
		// make the changes visible
		this.setVisible(true);
	}
	
	// this function looks at which check boxes were ticked and creates an
	// InvoiceEntryFormat object to match the user's specifications
	public InvoiceEntryFormat getUserInvoiceEntryFormat()
	{
		InvoiceEntryFormat retVal = new InvoiceEntryFormat();
		
		// set the various functions to state of their respective checkboxes
		retVal.setTitle(entryTitleCheckBox.isSelected());
		retVal.setDescription(entryDescriptionCheckBox.isSelected());
		retVal.setFormatedDateText(entryTimeStampCheckBox.isSelected());
		retVal.setLocation(entryLocationCheckBox.isSelected());
		retVal.setHours(entryTotalTimeCheckBox.isSelected());
		
		return retVal;
	}
	
	// this function uses the list of CalendarReaderButtons to create a list of all Calendars
	// being used in the creation of this Invoice. This list is displayed using the listOfCalendars
	// JPanel object
	public void updateCalendarList()
	{
		// create a new listOfCalendars to hold the results
		listOfCalendars.removeAll();
		listOfCalendars.setLayout(new BoxLayout(listOfCalendars, BoxLayout.Y_AXIS));
		
		// if there are calendars to list
		if(calendars.size() > 0)
		{
			// for every invoice in the list
			for(CalendarReader calendar : calendars)
			{
				// we create a new CalendarReaderButton using the current Calendar Object in the loop
				// then add that Button to the panel of all Calendar Reader GUI elements
				listOfCalendars.add(new CalendarReaderButton(calendar));
			}
		}
		// if there are no calendars to list
		else
		{
			// re-create the original JLabel that explains there are no Calendars to draw
			listOfCalendars.add(new JLabel("-- NONE --"));
		}
		
		// make the changes visible
		this.setVisible(true);
	}

	// this function is called by CalendarReaderButtons to remove themselves from the list
	public void removeCalendar(CalendarReader toRemove) throws DataNotLoaded
	{
		// for every Invoice Entry Button currently loaded
		for(int i=0; i<invoicePanel.getComponentCount(); i++)
		{
			InvoiceEntryButton entry = (InvoiceEntryButton)invoicePanel.getComponent(i);
			
			// if this entry is selected AND is in the Calendar that is being removed
			if(entry.isSelected() && entry.getEntry().getParent().equals(toRemove))
			{
				// we remove it's hours from the total
				InvoiceCreatorGUI.getInstance().subHours(entry.getEntry());
			}
		}
		// remove the specified CalendarReaderButton
		calendars.remove(toRemove);
		
		// reload all the invoices
		loadInvoices();
		
		// update the user's list of CalendarReaderButtons
		updateCalendarList();
	}
	
	// this function makes all the invoice entries from the specified Calendar either visible
	// or invisible depending on the flag
	public void makeVisible(CalendarReader calendar, boolean flag)
	{
		// for every invoice,
		for(int i=0; i<invoices.size(); i++)
		{
			// get this invoice entry button
			InvoiceEntryButton thisEntry = (InvoiceEntryButton)invoicePanel.getComponent(i);
			
			// thisEntry is part of the calendar being marked invisible
			if(thisEntry.getEntry().getParent().equals(calendar))
			{
				// set this Entry to the specified value
				thisEntry.makeVisible(flag);
			}
		}	
	}
	
	// this function makes all the invoice entries within the date Range either visible
	// or invisible depending on the flag
	public void onlyVisibileBetween(Date start, Date end)
	{
		// for every invoice,
		for(int i=0; i<invoices.size(); i++)
		{
			// get this invoice entry button
			InvoiceEntryButton thisEntry = (InvoiceEntryButton)invoicePanel.getComponent(i);
			
			// thisEntry is part of the calendar being marked invisible
			if( (thisEntry.getEntry().getStartDate().compareTo(start)>=0) 
				 && (thisEntry.getEntry().getEndDate().compareTo(end)<=0) )
			{
				// set this Entry to visible
				thisEntry.makeVisible(true);
			}
			else // the entry does not fall within the specified range
			{
				// set this Entry to hidden
				thisEntry.makeVisible(false);
			}
		}
	}
	
	public void setSelectionForVisible(boolean flag)
	{
		// for every invoice,
		for(int i=0; i<invoices.size(); i++)
		{
			// get this invoice entry button
			InvoiceEntryButton thisEntry = (InvoiceEntryButton)invoicePanel.getComponent(i);
			
			// if this entry is visible
			if(thisEntry.getVisibility())
			{
				// set this Entry to whatever the flag is set to
				thisEntry.setSelection(flag);
			}
		}
	}

	public void setPageCode(String code) 
	{
		customPageCode = code;
	}
	
	public String getPageCode()
	{
		return customPageCode;
	}
	
	public boolean isPageSet()
	{
		return !(customPageCode == null);
	}

	@Override
	public void focusGained(FocusEvent e) 
	{
		JTextComponent object = (JTextComponent)e.getSource();
		object.setText("");
		object.removeFocusListener(this);
	}

	@Override
	public void focusLost(FocusEvent e) 
	{
		// do nothing
	}
	
	// this function adds the hours of the passed InvoiceEntry (thisEntry)
	// to the running total for easy reference
	public void addHours(InvoiceEntry thisEntry)
	{
		float hours = Float.parseFloat(invoiceInterfaceHours.getText());
		hours = hours + thisEntry.getHours();
		invoiceInterfaceHours.setText(Float.toString(hours));
	}

	// this function subtracts the hours of the passed InvoiceEntry (thisEntry)
	// from the running total for easy reference
	public void subHours(InvoiceEntry thisEntry)
	{
		float hours = Float.parseFloat(invoiceInterfaceHours.getText());
		hours = hours - thisEntry.getHours();
		invoiceInterfaceHours.setText(Float.toString(hours));
	}
}
